@@ -1,8 +1,10 @@
-# This module contains the base objects needed
+"""
+This module contains the base objects of the JSON structure
+"""
 import json
-import os
 import sys
 from datetime import date, datetime
+from pathlib import Path
 from uuid import uuid4
 
 import requests
@@ -17,7 +19,7 @@ from jsonutils.exceptions import (
     JSONQueryException,
     JSONQueryMultipleValues,
 )
-from jsonutils.functions.decorators import catch_exceptions
+from jsonutils.functions.decorators import return_value_on_exception
 from jsonutils.functions.parsers import (
     _parse_html_table,
     _parse_query,
@@ -29,74 +31,126 @@ from jsonutils.functions.parsers import (
     parse_timestamp,
     url_validator,
 )
+from jsonutils.functions.seekers import (
+    _eval_object,
+    _json_from_path,
+    _relative_to,
+    _set_object,
+    empty,
+)
 from jsonutils.query import All, KeyQuerySet, ParentList, QuerySet
 from jsonutils.utils.dict import UUIDdict, ValuesDict
 from jsonutils.utils.retry import retry_function
 
-try:
-    DjangoQuerySet = sys.modules["django"].db.models.QuerySet
-except Exception:
-    DjangoQuerySet = type(None)
+
+# ---- external modules ----
+def DjangoQuerySet():
+
+    try:
+        return sys.modules["django"].db.models.QuerySet
+    except Exception:
+        return type(None)
+
+
+def PandasDataFrame():
+
+    try:
+        return sys.modules["pandas"].DataFrame
+    except Exception:
+        return type(None)
+
+
+# ----------------------
 
 
 class JSONPath:
     """
     Object representing a JSON path for a given JSON object.
     Don't instanciate it directly.
+
+    Attributes
+    ----------
+        data: a pretty string representation of the path. Ex: 'data/0/name'.
+        expr: a python string representation of the path. Ex: '["data"][0]["name"]'.
+        keys: a tuple representation of the path. Ex: ("data", 0, "name").
     """
 
-    def __new__(cls, s=""):
-        obj = super().__new__(cls)
-        obj._string = s  # pretty path
-        obj._path = ""  # python json path
-        obj._keys = []  # list of dict keys
-        return obj
+    @staticmethod
+    def _cast_to_int(s):
+        if s.isdigit():
+            return int(s)
+        else:
+            return s
+
+    def __init__(self, path=()):
+        if isinstance(
+            path, str
+        ):  # if path is given as an string like 'A/0/B', cast it to a tuple like ("A", 0, "B")
+            path = tuple(self._cast_to_int(i) for i in path.split("/") if i != "")
+        if not isinstance(path, tuple):
+            raise TypeError(
+                f"Argument 'path' must be a tuple or str instance, not {type(path)}"
+            )
+        self._keys = path  # tuple of dict keys
 
     @property
     def data(self):
-        return self._string
+        return "/".join(map(str, self._keys))
 
     @property
     def expr(self):
-        return self._path
+        output = ""
+        for k in self._keys:
+            if isinstance(k, str):
+                output += f'["{k}"]'
+            elif isinstance(k, int):
+                output += f"[{k}]"
+        return output
 
     @property
     def keys(self):
-        return list(reversed(self._keys))
+        """Returns a tuple with the object's path keys"""
+        return self._keys
 
-    def relative_to(self, child):
-        """Calculate jsonpath relative to child's jsonpath"""
-        # TODO review this algorithm, because it fails when a key has /
+    def relative_to(self, node):
+        """
+        Calculate jsonpath relative to a parent node's jsonpath.
+        Examples of relative paths:
+            * self -> ("A", 0, "B", "C")
+                                            ==> ("B", "C")
+            * node -> ("A", 0)
+        Returns a new JSONPath instance
+        """
 
-        if not isinstance(child, JSONNode):
+        if not isinstance(node, JSONNode):
             raise TypeError(
-                f"child argument must be a JSONNode instance, not {type(child)}"
+                f"child argument must be a JSONNode instance, not {type(node)}"
             )
-        if child.jsonpath._path:
-            root = child.jsonpath._path
-        else:
-            root = ""
-        full_path = self._path
 
-        common_prefix = os.path.commonprefix([root, full_path])
+        self_path = self._keys
+        other_path = node.jsonpath._keys
 
-        return full_path.replace(common_prefix, "")
+        result = _relative_to(self_path, other_path)
+        jsonpath = JSONPath(result)
+        return jsonpath
 
     def _update(self, **kwargs):
         if (key := kwargs.get("key")) is not None:
-            self._string = str(key) + "/" + self._string
-            self._path = f'["{key}"]' + self._path
-            self._keys.append(key)
+            self._keys = (key,) + self._keys
         elif (index := kwargs.get("index")) is not None:
-            self._string = str(index) + "/" + self._string
-            self._path = f"[{index}]" + self._path
-            self._keys.append(index)
+            self._keys = (index,) + self._keys
 
     def __eq__(self, other):
-        return (self._string == other) or (self._path == other)
+        if isinstance(other, (tuple, list)):
+            return self._keys == tuple(other)
+        if isinstance(other, str):
+            return self.data == other
+        if isinstance(other, self.__class__):
+            return self._keys == other._keys
+        return NotImplemented
 
     def __repr__(self):
-        return self._string
+        return self.data
 
 
 class JSONObject:
@@ -137,8 +191,6 @@ class JSONObject:
             return JSONNull(data)
         elif isinstance(data, dict):
             return JSONDict(data)
-        elif isinstance(data, (list, tuple, DjangoQuerySet)):
-            return JSONList(data)
         elif isinstance(data, bool):
             return JSONBool(data)
         elif isinstance(data, str):
@@ -149,6 +201,11 @@ class JSONObject:
             return JSONFloat(data)
         elif isinstance(data, int):
             return JSONInt(data)
+        # data from external libraries
+        elif isinstance(data, (list, tuple, DjangoQuerySet())):
+            return JSONList(data)
+        elif isinstance(data, PandasDataFrame()):
+            return JSONDict(json.loads(data.to_json()))
         else:
             if not raise_exception:
                 return JSONUnknown(data)
@@ -201,6 +258,7 @@ class JSONObject:
     @staticmethod
     def read_html_table(
         data,
+        parser="lxml",
         raise_exception=True,
         attrs={},
         recursive=True,
@@ -208,8 +266,31 @@ class JSONObject:
         link_prefix=None,
         **kwargs,
     ):
+        """
+        Loads an html table from an url or string as node instance.
+
+        Params
+        ------
+            data: an string representing the raw data. If an url is passed,
+                  then it will try to make a request to the web and locate a valid html table.
+                  If, on the other hand, the text string represents an html structure,
+                  it will act on it without making any request.
+            parser: parse library to be used by `BeautifulSoup`.
+            raise_exception: if True, then any errors derived from the table lookup will be thrown as response.
+                             If False, it will return None on errors.
+            attrs: a dictionary to be passed as an argument to the `find` function of `BeautifulSoup` object.
+                   It is useful, for example, if you want to specify a particular attribute that must be present in the HTML tags of the table.
+            recursive: it will be passed as an argument to the `find` function of `BeautifulSoup` object.
+                       Specifies whether the table lookup should be recursive or not.
+            parse_links: if True, then instead of taking the text inside a tag, it will retrieve the first href link, if it exists.
+            link_prefix: when `parse_links` is selected, we can prepend an string to href links.
+            kwargs: extra arguments to be passed to the request, such as headers, proxies, etc.
+        """
+        # TODO needs a test
         if not isinstance(data, str):
-            raise TypeError(f"Argument data must be an str instance")
+            raise TypeError(
+                f"Argument data must be an 'str' instance, not {type(data)}"
+            )
 
         if url_validator(data):
             try:
@@ -220,7 +301,7 @@ class JSONObject:
                 else:
                     return
 
-            soup = BeautifulSoup(req.content, "lxml")
+            soup = BeautifulSoup(req.content, parser)
             table = soup.find("table", attrs, recursive)
 
             if not table:
@@ -246,7 +327,7 @@ class JSONObject:
             return result
 
         else:  # data is already an html string
-            soup = BeautifulSoup(data, "lxml")
+            soup = BeautifulSoup(data, parser)
             table = soup.find("table", attrs, recursive)
 
             if not table:
@@ -271,6 +352,15 @@ class JSONObject:
 
             return result
 
+    @classmethod
+    def from_path(cls, iterable):
+        """
+        Build a JSONObject from a list of leaf node paths
+        """
+
+        obj = _json_from_path(iterable)
+        return cls(obj)
+
 
 class JSONNode:
     """
@@ -303,6 +393,15 @@ class JSONNode:
     def json_encode(self, **kwargs):
         return json.dumps(self, cls=JSONObjectEncoder, **kwargs)
 
+    @property
+    def is_leaf(self):
+        """Check if this node is a leaf node (no childs)"""
+        try:
+            child_number = self._child_objects.__len__()
+        except AttributeError:
+            return True
+        return not bool(child_number)
+
     @property
     def json_decode(self):
         return json.loads(json.dumps(self, cls=JSONObjectEncoder))
@@ -340,6 +439,27 @@ class JSONNode:
             parent = parent.parent
         return last
 
+    def update(self, new_obj):
+        """
+        Update this node within JSONObject from which it is derived
+        """
+
+        is_callable = callable(new_obj)
+
+        path = self.jsonpath
+        root = self.root
+
+        if is_callable:
+            try:
+                root.set_path(path, new_obj(root.eval_path(path)))
+            except Exception:
+                return False
+            else:
+                return True
+        else:
+            root.set_path(path, new_obj)
+        return True
+
     def values(self, *keys, search_upwards=True, flat=False, **kwargs):
         if flat is True and len(keys) > 1:
             raise ValueError(
@@ -400,7 +520,7 @@ class JSONNode:
         return output_dict if not flat else list(output_dict.values())[0]
 
     def __str__(self):
-        return self.json_encode(indent=4)
+        return self.json_encode(indent=4, ensure_ascii=False)
 
     # ---- ACTION METHODS ----
     def gt_action(self, other):
@@ -503,6 +623,11 @@ class JSONNode:
 
         return _apply(self, other)
 
+    def bool_action(self, other):
+        from jsonutils.functions.actions import _bool
+
+        return _bool(self, other)
+
 
 class JSONCompose(JSONNode):
     """
@@ -535,6 +660,60 @@ class JSONCompose(JSONNode):
             for index, item in enumerate(self):
                 self.__setitem__(index, item)
 
+    def path_exists(self, iterable):
+        """
+        Returns a boolean especifing if selected path exist in the composed object.
+        """
+        from jsonutils.functions.seekers import _eval_object
+
+        if isinstance(iterable, (str, int)):
+            iterable = (iterable,)
+        elif isinstance(iterable, JSONPath):
+            iterable = iterable.keys
+
+        try:
+            _eval_object(self, iterable)
+        except (IndexError, KeyError):
+            return False
+        return True
+
+    def set_path(self, path, value):
+        """
+        Set composed object's value on iterable path
+        Example
+        -------
+
+        data = JSONObject(
+            {
+                "data": [
+                    {
+                        "A": True
+                    }
+                ]
+            }
+        )
+
+        >> data.set_path(
+            ("data",0,"B"),
+            False
+        )
+
+        >> data
+            {
+                "data": [
+                    {
+                        "A": True,
+                        "B": False
+                    }
+                ]
+            }
+        """
+
+        if isinstance(path, JSONPath):
+            path = path.keys
+
+        return _set_object(self, path, value)
+
     def query(
         self,
         recursive_=None,
@@ -758,21 +937,50 @@ class JSONCompose(JSONNode):
                 if item.is_composed and recursive:
                     item._remove_annotations()
 
-    @catch_exceptions
-    def _eval_path(self, path, fail_silently=False):
+    @return_value_on_exception(empty, (IndexError, KeyError))
+    def eval_path(
+        self, path, fail_silently=False, native_types_=None, value_on_exception=None
+    ):
         """
-        Evaluate JSONCompose object over a jsonpath
+        Evaluate JSONCompose object over a jsonpath.
+
+        Arguments
+        ---------
+            path: nested path on which the object will be evaluated.
+            fail_silently: if True, it will return empty in case of errors (missing paths).
+            native_types_: if True, then the result will be a Python object whenever a right path is found,
+                           instead of a JSONNode object.
+            value_on_exception: the value which will be returned on errors, if fail_silently is True (by default, empty)
         """
-        # TODO add test
-        if isinstance(path, str):
-            aux = JSONPath()
-            aux._path = path
-            path = aux
-        return eval(f"self{path.expr}")
+        if isinstance(path, JSONPath):
+            path = path.keys
+        elif isinstance(path, str):
+            if "/" in path:
+                path = tuple(JSONPath._cast_to_int(i) for i in path.split("/") if i)
+            else:
+                path = (path,)
+
+        if native_types_ is None:
+            native_types_ = config.native_types
+
+        if not isinstance(path, (tuple, list)):
+            raise TypeError(
+                f"path argument must be a JSONPath, str, tuple or list instance, not {type(path)}"
+            )
+
+        if native_types_:
+            res = _eval_object(self, path)._data
+        else:
+            res = _eval_object(self, path)
+        return res
 
     def traverse_json(self):
         """
-        Traverse recursively over all json data
+        Traverse recursively over all json data.
+
+        Returns
+        -------
+            QuerySet object
         """
 
         output_list = QuerySet()
@@ -781,13 +989,53 @@ class JSONCompose(JSONNode):
         for child in children:
             serialized_child = child._data
             output_list.append(
-                JSONObject({"path": child.jsonpath.expr, "value": serialized_child})
+                JSONObject({"path": child.jsonpath.keys, "value": serialized_child})
             )
             if child.is_composed:
                 output_list += child.traverse_json()
 
         return output_list
 
+    def to_path(self):
+        """
+        Retrieve a dict of json leaf nodes paths
+
+        Example
+        -------
+
+        data = JSONObject(
+            {
+                "A": [
+                    {
+                        "A1": 1
+                    },
+                    {
+                        "A2": 2
+                    }
+                ],
+                "B": 3
+            }
+        )
+
+        >> data.to_path()
+            {
+                ('A', 0, 'A1'): 1,
+                ('A', 1, 'A2'): 2,
+                ('B',): 3
+            }
+        """
+        output_dict = {}
+
+        children = self._child_objects.values()
+        for child in children:
+            if child.is_leaf:
+                serialized_child = child._data
+                output_dict[child.jsonpath.keys] = serialized_child
+            if child.is_composed:
+                output_dict.update(child.to_path())
+
+        return output_dict
+
     def check_valid_types(self):
         """Check if json object has valid types (not unknown types)"""
 
@@ -809,6 +1057,61 @@ class JSONCompose(JSONNode):
 
         return _validate_data(self, schema)
 
+    def save(self, path, create_path=True, ensure_ascii=False, indent=4, **kwargs):
+        """
+        Save the JSON Composed object to a file.
+        Arguments
+        ---------
+            path: full path where to store the output file
+            create_path: if True, then path to file will be created if doesn't exist
+            ensure_ascii: if you want to handle unicode values
+            indent: number of indents (default 4)
+        """
+        _path = Path(path).resolve()
+
+        if create_path is True:
+            _path.parent.mkdir(parents=True, exist_ok=True)
+
+        with open(path, "w") as file:
+            json.dump(
+                self,
+                file,
+                cls=JSONObjectEncoder,
+                ensure_ascii=ensure_ascii,
+                indent=indent,
+                **kwargs,
+            )
+
+    def merge(self, other, kind="inner_join"):
+        """
+        Makes a JSON merge with other JSON object
+        """
+        # TODO
+        other = JSONObject(other)
+        left_paths = set(self.to_path().keys())
+        right_paths = set(other.to_path().keys())
+
+        path_dict = {}
+
+        kind = kind.lower().strip()
+
+        if kind == "inner_join":
+            resulting_set = set.intersection(left_paths, right_paths)
+        elif kind == "outer_join":
+            resulting_set = set.union(left_paths, right_paths)
+        else:
+            raise ValueError(
+                f"Argument 'kind' must match one of the following options: 'inner_join', 'outer_join'"
+            )
+
+        for path in resulting_set:
+            left_value = self.eval_path(path, fail_silently=True, native_types_=True)
+            right_value = other.eval_path(path, fail_silently=True, native_types_=True)
+
+            path_dict[path] = right_value or left_value
+
+        return JSONObject.from_path(path_dict)
+
 
 class JSONSingleton(JSONNode):
     """
@@ -828,7 +1131,7 @@ class JSONSingleton(JSONNode):
 
 # ---- COMPOSE OBJECTS ----
 class JSONDict(dict, JSONCompose):
-    """ """
+    """A Dict object"""
 
     _DEFAULT = object()
     get = JSONCompose.get  # override get method
@@ -893,6 +1196,33 @@ class JSONDict(dict, JSONCompose):
         obj = cls(self.json_decode)
         return obj
 
+    def get_fields(self, *args, inplace=False):
+        """
+        Returns a subset of this dict with selected fields, if it exists
+        """
+        # TODO add test
+        if inplace:
+            for key in list(self):
+                if key not in args:
+                    self.pop(key)
+            return
+
+        obj = {k: v for k, v in self.items() if k in args}
+        return JSONDict(obj)
+
+    def get_fields_except(self, *args, inplace=False):
+        """
+        Returns those fields from the dict that are not contained in *args
+        """
+        if inplace:
+            for key in list(self):
+                if key in args:
+                    self.pop(key)
+            return
+
+        obj = {k: v for k, v in self.items() if k not in args}
+        return JSONDict(obj)
+
     def pop(self, key, default=_DEFAULT):
         """
         When removing a key from dict, we also must unregister the corresponding child
@@ -906,10 +1236,46 @@ class JSONDict(dict, JSONCompose):
         else:
             return default
 
-    def to_django_model(self, model, fail_silently=False):
-        """Translates a JSON dict to Django model new instance"""
+    def to_django_model(
+        self,
+        model,
+        manager="objects",
+        kind="create",
+        get_fields=(),
+        default_fields=(),
+        fail_silently=False,
+    ):
+        """
+        Translates a JSON dict to Django model new instance
+        Arguments
+        ---------
+            manager:
+            kind:
+            get_fields: the subset of fields which will be gathered in the get request
+        """
 
-        return _to_django_model(self, model, fail_silently=fail_silently)
+        def _check_type(arg):
+            if isinstance(arg, str):
+                return (arg,)
+            elif isinstance(arg, (list, tuple)):
+                return arg
+            elif isinstance(arg, type(None)):
+                return ()
+            else:
+                raise TypeError(f"Argument '{arg}' must be an iterable instance")
+
+        get_fields = _check_type(get_fields)
+        default_fields = _check_type(default_fields)
+
+        return _to_django_model(
+            self,
+            model,
+            manager=manager,
+            kind=kind,
+            get_fields=get_fields,
+            default_fields=default_fields,
+            fail_silently=fail_silently,
+        )
 
     # ---- COMPARISON METHODS ----
     def __eq__(self, other):
@@ -929,7 +1295,7 @@ class JSONDict(dict, JSONCompose):
 
 
 class JSONList(list, JSONCompose):
-    """ """
+    """A list object"""
 
     def __init__(self, *args, **kwargs):
 
@@ -1332,6 +1698,9 @@ class JSONBool(JSONSingleton):
     def __repr__(self):
         return str(self._data)
 
+    def __str__(self):
+        return self.__repr__()
+
     def __bool__(self):
         return self._data
 
@@ -1372,6 +1741,9 @@ class JSONNull(JSONSingleton):
     def __repr__(self):
         return "None"
 
+    def __str__(self):
+        return self.__repr__()
+
     def __bool__(self):
         return False
 
