@@ -3,6 +3,7 @@
 import ast
 import re
 from datetime import date, datetime
+from functools import reduce
 from json import JSONDecoder
 
 import jsonutils.base as base
@@ -398,6 +399,7 @@ def parse_datetime(
     only_date=False,
     fail_silently=False,
     return_string=False,
+    tzone="utc",
 ):
     """
     If only_check is True, then this algorithm will just check if string s matchs a datetime format (no errors).
@@ -431,13 +433,13 @@ def parse_datetime(
                 )
             else:  # if not tzinfo is shown, put utc as default
                 return (
-                    unified_datetime.replace(tzinfo=pytz.utc)
+                    unified_datetime.replace(tzinfo=pytz.timezone(tzone))
                     if not only_date
                     else datetime(
                         unified_datetime.year,
                         unified_datetime.month,
                         unified_datetime.day,
-                        tzinfo=pytz.utc,
+                        tzinfo=pytz.timezone(tzone),
                     )
                 )
         else:
@@ -508,11 +510,14 @@ def parse_datetime(
             min = group_numbers.get("min")
             sec = group_numbers.get("sec")
 
-            off_sign = groups.get("off_sign") or "+"
-            off_hh = groups.get("off_hh") or "00"
-            off_mm = groups.get("off_mm") or "00"
+            off_sign = groups.get("off_sign")
+            off_hh = groups.get("off_hh")
+            off_mm = groups.get("off_mm")
 
-            tzone = datetime.strptime(f"{off_sign}{off_hh}:{off_mm}", "%z").tzinfo
+            if all((off_sign, off_hh, off_mm)):
+                tzone = datetime.strptime(f"{off_sign}{off_hh}:{off_mm}", "%z").tzinfo
+            else:
+                tzone = pytz.timezone(tzone)
 
             try:
                 if only_date:
@@ -630,10 +635,41 @@ def parse_http_url(url, fail_silently=False):
 
 
 @catch_exceptions
-def _to_django_model(node, model, fail_silently=False):
+def _to_django_model(
+    node,
+    model,
+    manager="objects",
+    kind="create",
+    get_fields=(),
+    default_fields=(),
+    fail_silently=False,
+):
     # TODO add more checks
-    data = node.json_decode
-    obj, _ = model.objects.get_or_create(**data)
+    # TODO add test
+
+    if get_fields:
+        data = node.get_fields(*get_fields).json_decode
+        if default_fields:
+            defaults = node.get_fields(*get_fields).json_decode
+        else:
+            defaults = node.get_fields_except(*get_fields).json_decode
+    else:
+        data = node.json_decode
+        if default_fields:
+            defaults = node.get_fields(*get_fields).json_decode
+        else:
+            defaults = data
+
+    if kind == "create":
+        method = "get_or_create"
+    elif kind == "update":
+        method = "update_or_create"
+    else:
+        raise ValueError(
+            f"Argument 'kind' must be one of the following: 'create', 'update'. Not {kind}"
+        )
+    full_function = reduce(getattr, [manager, method], model)
+    obj, _ = full_function(**data, defaults=defaults)
 
     return obj
 
@@ -747,15 +783,6 @@ def url_validator(url, public=False, return_match=False, optative_protocol=False
 
 
 def _parse_html_table(table, parse_links, link_prefix):
-    """
-    Parses an html table
-
-    Arguments
-    ---------
-        table: a BeautifulSoup object which contains a table tag
-        parse_links: if True, then instead of taking the text inside a tag, it will retrieve the first href link, if it exists
-        link_prefix: when parse_links is selected, we can prepend an string to href links
-    """
 
     # types check
 
