@@ -23,6 +23,7 @@ from jsonutils.base import (
 )
 from jsonutils.encoders import JSONObjectEncoder
 from jsonutils.exceptions import JSONQueryException, JSONQueryMultipleValues
+from jsonutils.functions.seekers import empty
 from jsonutils.query import All, ExtractYear, QuerySet, SingleQuery, ValuesList
 
 BASE_PATH = Path(js.__file__).parent.resolve()
@@ -122,7 +123,7 @@ class JsonTest(unittest.TestCase):
         self.assertIsInstance(test2.data, JSONUnknown)
         self.assertEqual(
             test2.check_valid_types(),
-            (False, [{"data": {"path": "data/", "type": type}}]),
+            (False, [{"data": {"path": "data", "type": type}}]),
         )
 
     def test_types(self):
@@ -166,7 +167,7 @@ class JsonTest(unittest.TestCase):
 
         self.assertEqual(test2[-1], test1._1.Dict.List)
         self.assertNotEqual(test2[-1]._0.jsonpath, test1._1.Dict.List._0.jsonpath)
-        self.assertEqual(test2[-1]._0.jsonpath, "3/0/")
+        self.assertEqual(test2[-1]._0.jsonpath, "3/0")
 
         test2.append(test1._1.Dict.List, serialize_nodes=False)
         # if not serializing nodes, the new appended jsonpath will not match the path within test2 jsonobject
@@ -683,7 +684,40 @@ class JsonTest(unittest.TestCase):
         self.assertTrue(q5._check_against_node(test._1.date))
         self.assertTrue(q6._check_against_node(test._0.B))
 
-    def test_update(self):
+    def test_update_in_nodes(self):
+        test = JSONObject(
+            {
+                "data": [
+                    {"name": "Dan", "id": [1, 2]},
+                    {
+                        "name": "Don",
+                        "id": [0, 3],
+                        "comments": [{"id": 1, "value": "Great"}],
+                    },
+                ],
+                "timestamp": "2021-01-01 10:00:00Z",
+                "valuation": {"pre": 10, "post": 9},
+            }
+        )
+        # update timestamp
+        pre_path = test.get(timestamp__year=2021).jsonpath
+
+        self.assertEqual(
+            test.get(timestamp__year=2021).update(lambda x: "2022" + x[4:]), True
+        )
+        self.assertEqual(test.get(timestamp__year=2022), "2022-01-01 10:00:00Z")
+        self.assertEqual(test.get(timestamp__year=2021, throw_exceptions_=False), None)
+        self.assertEqual(pre_path, test.timestamp.jsonpath)
+
+        # failing updating timestamp
+
+        self.assertEqual(test.timestamp.update(lambda x: x ** 2), False)
+        self.assertEqual(test.get(timestamp__year=2022), "2022-01-01 10:00:00Z")
+
+        # update id
+        self.assertEqual(test.get(id__0=1).update(lambda x: x + ["fake"]), True)
+
+    def test_update_in_queries(self):
 
         test = JSONObject(
             {
@@ -773,7 +807,7 @@ class JsonTest(unittest.TestCase):
         )
         self.assertEqual(test2.query(key=All), [111])
         self.assertListEqual(test2.query(index=All), [333, 222])
-        self.assertEqual(test2.query(index=333).first().jsonpath, "nested/index/")
+        self.assertEqual(test2.query(index=333).first().jsonpath, "nested/index")
 
         self.assertTrue(test3.query(A__contains=4).exists())
 
@@ -805,7 +839,7 @@ class JsonTest(unittest.TestCase):
             ),
         )
 
-        self.assertEqual(test._1.Dict.a2.jsonpath, "1/Dict/a2/")
+        self.assertEqual(test._1.Dict.a2.jsonpath, "1/Dict/a2")
         self.assertEqual(test.query(a1=1).count(), 3)
         self.assertNotEqual(test, self.test1)
 
@@ -861,7 +895,7 @@ class JsonTest(unittest.TestCase):
         self.assertEqual(test.query(A__parents__c_key__1__contains="ey").count(), 1)
         self.assertEqual(
             test.query(A__parents__c_key__1__contains="ey").first().jsonpath,
-            "root/root_dict/child/A/",
+            "root/root_dict/child/A",
         )
         self.assertEqual(
             test.get(A__parents__0=0, A__parents__index=1, throw_exceptions_=True), 1
@@ -870,7 +904,7 @@ class JsonTest(unittest.TestCase):
             test.get(
                 A__parents__0=0, A__parents__index=1, throw_exceptions_=True
             ).jsonpath,
-            "root/root_list/1/child/A/",
+            "root/root_list/1/child/A",
         )
         self.assertIsNone(
             test.get(
@@ -1086,12 +1120,80 @@ class JsonTest(unittest.TestCase):
         self.assertListEqual(
             test.traverse_json().values("path", flat=True),
             [
-                "[0]",
-                '[0]["A"]',
-                '[0]["B"]',
-                '[0]["B"]["B1"]',
-                '[0]["B"]["B2"]',
-                '[0]["B"]["B2"][0]',
-                '[0]["B"]["B2"][1]',
+                [
+                    0,
+                ],
+                [0, "A"],
+                [0, "B"],
+                [0, "B", "B1"],
+                [0, "B", "B2"],
+                [0, "B", "B2", 0],
+                [0, "B", "B2", 1],
             ],
         )
+
+    def test_json_paths(self):
+        test = JSONObject({"A": {"B": [{"C": {"D": 1}}]}})
+
+        self.assertEqual(test.A.B._0.C.D.jsonpath.keys, ("A", "B", 0, "C", "D"))
+        self.assertEqual(test.A.B._0.C.jsonpath.expr, '["A"]["B"][0]["C"]')
+
+        test.set_path(("A", "B", 0, "D"), 10)
+        self.assertTupleEqual(
+            test.query_key("*", exact=10).first().jsonpath.keys, ("A", "B", 0, "D")
+        )
+        test.set_path(("B",), 20)
+        self.assertTupleEqual(
+            test.query_key("*", exact=20).first().jsonpath.keys, ("B",)
+        )
+
+        self.assertRaises(TypeError, lambda: test.set_path(("A", "B", "A"), 30))
+
+    def test_path_exists(self):
+        test = JSONObject({"A": {"B": [{"C": {"D": 1}}]}})
+
+        self.assertTrue(test.path_exists(("A",)))
+        self.assertTrue(test.path_exists(("A", "B")))
+        self.assertTrue(test.path_exists(("A", "B", 0)))
+        self.assertTrue(test.path_exists(("A", "B", 0, "C")))
+        self.assertTrue(test.path_exists(("A", "B", 0, "C", "D")))
+
+        self.assertFalse(test.path_exists(("a",)))
+        self.assertFalse(test.path_exists(("A", 0)))
+        self.assertFalse(test.path_exists(("A", "B", 1)))
+        self.assertFalse(test.path_exists(("A", "B", 0, "c")))
+        self.assertFalse(test.path_exists(("A", "B", 0, "C", "d")))
+
+    def test_eval_path(self):
+        test = JSONObject({"A": {"B": [{"C": 1}]}})
+
+        self.assertTrue(test.eval_path(("A",)), {"B": [{"C": 1}]})
+        self.assertTrue(test.eval_path("A", {"B": [{"C": 1}]}))
+        self.assertTrue(test.eval_path(("A", "B")), [{"C": 1}])
+        self.assertTrue(test.eval_path(("A", "B", 0)), {"C": 1})
+        self.assertTrue(test.eval_path(("A", "B", 0, "C")), 1)
+        self.assertEqual(test.eval_path(("A", "B", 0, "C")).jsonpath, "A/B/0/C")
+        self.assertRaises(
+            AttributeError,  # jsonpath attribute should not exist when returning native types
+            lambda: test.eval_path(("A", "B", 0, "C"), native_types_=True).jsonpath,
+        )
+
+        # ---- FAIL PATHS ----
+        self.assertRaises(KeyError, lambda: test.eval_path(("B",)))
+        self.assertRaises(KeyError, lambda: test.eval_path(("A", "C")))
+        self.assertRaises(IndexError, lambda: test.eval_path(("A", "B", 1)))
+        self.assertEqual(
+            test.eval_path(("A", "A"), fail_silently=True), empty
+        )  # not raises an error if fail_silently is True
+        self.assertEqual(
+            test.eval_path(("A", "A"), fail_silently=True, value_on_exception=99), 99
+        )
+        self.assertEqual(
+            test.eval_path(("A", "A"), fail_silently=True, value_on_exception=0), 0
+        )
+        self.assertEqual(
+            test.eval_path((0, 0), fail_silently=True, value_on_exception=""), ""
+        )
+        self.assertEqual(
+            test.eval_path((0, 0), fail_silently=True, value_on_exception=None), None
+        )
